/*
 * Quick benchmark program. Uses single and double precision floats to
 * heat up the CPU. It also checks how fast your CPU can copy memory from
 * one place to another.
 *
 * Note: This is quick and dirty.
 *
 * Copyright (C) 2007 Joris van Rooij <jorrizza@jrrzz.net>
 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>
#include <float.h>

/*
 * Get timestamp in milliseconds.
 */
unsigned long int get_milliseconds (void) {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

/*
 * REALLY nasty way to disply CPu type.
 * You know what; no one gives a shit!
 */
void print_cpu_type (void) {
  FILE *f;
  char buf[400], name[400], *found;
  unsigned int i;

  if (!(f = fopen("/proc/cpuinfo", "r"))) {
    printf("Unknown CPU\n");
    return;
  }
  fgets(buf, 400, f);
  while (!feof(f)) {
    buf[399] = '\0';
    buf[398] = '\n';
    found = strstr(buf, "model name");
    if (found != NULL) {
      for (i = 0; i < strlen(found); i++) {
        if (found[i] == ':') {
          strncpy(name, found + i + 2, strlen(found) - i - 2);
          printf("%sRunning %d-bit\n", name, (int) sizeof(size_t) * 8);
          fclose(f);
          return;
        }
      }
    }
    fgets(buf, 400, f);
  }
  fclose(f);
}

int main (int argc, char *argv[]) {
  /*
   * Initial stuff.
   * The third value is for simulating the after-calculus memory transfer
   * that usually occurs in real situations.
   */
  unsigned long int begin_bench, begin_test;
  int i1, i2, i3;
  float f1, f2, f3;
  double d1, d2, d3;
  void *data1 = NULL, *data2 = NULL;
  unsigned long int data_count = 0;
  unsigned int data_len = 10485760;
  unsigned int i;
  unsigned int data_offset = 0;
  
  d1 = 666.666;
  f1 = 666.666;
  i1 = 666;
  d2 = DBL_MAX;
  f2 = FLT_MAX;
  i2 = 4294967295;
  
  printf("\nCPU Benchmark\n");
  print_cpu_type();
  
  data1 = malloc(data_len);
  data2 = malloc(data_len);
  if (data1 == NULL || data2 == NULL) {
    fprintf(stderr, "Could not register 20MB of memory\n");
    return 1;
  }
  
  /*
   * Start the benchmark.
   */
  begin_bench = get_milliseconds();
  
  /*
   * Hammer the memory for ten seconds and check how much we can hammer it.
   */
  printf("\nRunning memory throughput (10s 1M block-block copy) benchmark:\n");
  
  while (get_milliseconds() - begin_bench < 10000) {
    if (data_offset == data_len)
      data_offset = 0;
    memcpy(data1 + data_offset, data2 + data_offset, 1048576);
    data_offset += 1048576;
    data_count += 1048576;
  }
  
  printf("%ld bytes/sec\n", data_count / 10);
  
  /*
   * Get the integers flowing!
   * Not really interesting. GHz matters here.
   */
  printf("\nRunning integer (low math) benchmark:\n");
  begin_test = get_milliseconds();
  
  for (i = 0; i < 150000000; i++) {
    i3 = i1 + i2;
    i3 = i1 - i2;
    i3 = i1 * i2;
    i3 = i1 / i2;
  }
  
  printf("%ld milliseconds\n", get_milliseconds() - begin_test);
  
  /*
   * First float test. Single precision.
   * The Amiga will fail.
   */
  printf("\nRunning 32-bit single precision float (high math) benchmark:\n");
  begin_test = get_milliseconds();
  
  for (i = 0; i < 20000000; i++) {
    f3 = sin(f1);
    f3 = sin(f2);
    f3 = asin(f1);
    f3 = asin(f2);
    f3 = cos(f1);
    f3 = cos(f2);
    f3 = acos(f1);
    f3 = acos(f2);
    f3 = tan(f1);
    f3 = tan(f2);
    f3 = atan(f1);
    f3 = atan(f2);
  }
  
  printf("%ld milliseconds\n", get_milliseconds() - begin_test);

  /*
   * Second float test. Double precision.
   * Poor x86 machines.
   */
  printf("\nRunning 64-bit double precision float (high math) benchmark:\n");
  begin_test = get_milliseconds();
  
  for (i = 0; i < 20000000; i++) {
    d3 = sin(d1);
    d3 = sin(d2);
    d3 = asin(d1);
    d3 = asin(d2);
    d3 = cos(d1);
    d3 = cos(d2);
    d3 = acos(d1);
    d3 = acos(d2);
    d3 = tan(d1);
    d3 = tan(d2);
    d3 = atan(d1);
    d3 = atan(d2);
  }
  
  printf("%ld milliseconds\n", get_milliseconds() - begin_test);

  /*
   * Finish it and show results.
   */
  printf("\nBenchmark finished. Your CPU was raped for\n==[%d]==\nMilliseconds. Lower is better.\n", (unsigned int) (get_milliseconds() - begin_bench));
  
  free(data1);
  free(data2);
  
  return 0;
}
